unit DBFile;

interface

uses Objects;

type
  KeyStr = string[10];

  PField = ^TField;
  TField = record                      { ᠭ    (1-32) }
    Name: array[0..10] of Char;        {    10 ᨬ  (1-10) }
    Flag: Char;                        {   [C, N, D, L, M] (12) }
    Addr: Longint;                     {   砫   (13-14) }
    Size: Byte;                        {         (17) }
    Dec:  Byte;                        { ᫮    (18) }
    Filler: array[19..32] of Byte;     { -⥫    (19-32) }
  end;

  PBase = ^TBase;
  TBase = object(TObject)
    Dbf_Id: Byte;                      { 䨪 ⨯ 䠩  (1) }
    LUpDate: record                    {     (2-4) }
      Year, Month, Day: Byte           { , ,  }
    end;
    LastRec: Longint;                  { ⢮ ᥩ  (5-8) }
    Header: Word;                      {         (9-10) }
    RecSize: Word;                     {      (11-12) }
    Filler: array[13..32] of Byte;     { -⥫    (13-32) }
    Fields: PCollection;               { ᠭ   }
    Buffer: PChar;                     {  ⥪饩  }
    RecNo: Longint;
    Eof: Boolean;
    constructor Init(FName: string);
    destructor Done; virtual;
    function At(Key: KeyStr): PField;  {ᠭ    }
    function Col(No: Integer): PField; {ᠭ   }
    function FCount: Integer;          {⢮  }
    function Alias: string;            {    䠩 }
    procedure Append; virtual;         {   }
    procedure Commit; virtual;         {⮫  }
    procedure Update;                  {  .dbf}
    procedure GoRec(No: Longint); virtual;
    procedure Put(Key: KeyStr; var Info: string);
    function Get(Key: KeyStr): string;
  private
    Dbf: File;
    Dbt: File;
    Modified: Boolean;                 {  ஢ }
    Updated: Boolean;                  {    }
  end;

  Date = record                        {   Y,M,D }
    Year, Month, Day: Word
  end;

procedure CtoD(C: string; var D: Date);
function DtoC(D: Date): string;
function DtoS(D: Date): string;

implementation  {==========================================================}

uses Dos, Strings;

const
  BlockLen = $20;
  DataBase = $03;       {     ⨯ Memo }
  MemoFile = $83;       {  , ᮤঠ  Memo }
  MarkBeg: Byte = $0D;  { 砫  ᥩ  䠩 DBF }
  MarkEnd: Byte = $1A;  {   ᥩ;   Memo }
  Delimiter = '-';      { -ࠧ⥫   }

type
  PFieldsCollection = ^TFieldsCollection;
  TFieldsCollection = object(TCollection)
    procedure FreeItem(Item: Pointer); virtual;
  end;

procedure TFieldsCollection.FreeItem(Item: Pointer);
begin
  FreeMem(Item, BlockLen);
end;

function Space(Len: Byte): string;
var
  St: string;
begin
  St := '';
  for Len := 1 to Len do St := St + ' ';
  Space := St;
end;

function MAlloc(Size: Word): Pointer;
var
  P: Pointer;
begin
  if Size < MaxAvail then
  begin
    GetMem(P, Size);
    FillChar(P^, Size, 0);
  end else P := nil;
  MAlloc := P;
end;

function Words(St: string; No: Integer): string;
const
  CharSet: set of Char = ['0'..'9', 'A'..'Z', '_', 'a'..'z',
                          ''..'', ''..'', ''..'', ''..''];
var
  I: Integer;
  V: string;
begin
  V := St;
  for I := 1 to No do
  begin
    while (V <> '') and not (V[1] in CharSet) do Delete(V, 1, 1);
    if I <> No then while (V <> '') and (V[1] in CharSet) do Delete(V, 1, 1);
  end;
  I := 1;
  if No > 0 then while (I <= Length(V)) and (V[I] in CharSet) do Inc(I);
  Words := Copy(V, 1, I - 1);
end;

procedure Today(var D: Date);
var
  DayOfWeek: Word;
begin
  GetDate(D.Year, D.Month, D.Day, DayOfWeek)
end;

function Within(X, Min, Max: Real): Boolean;
begin
  Within := (X >= Min) and (X <= Max)
end;

procedure CtoD(C: string; var D: Date);
type
  Arr = array[1..3] of Word;
const
  SizeOfMonth: array[1..12] of Byte = { ⢮    }
    (31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31);
var
  i, Code: Integer;
  Temp: Date;
begin
  for i := 1 to 3 do
  begin
    Arr(D)[Abs(i-4)] := 0;
    Val(Copy(Words(C, i), 1, 4), Arr(D)[Abs(i-4)], Code);
  end;
  Today(Temp);
  if D.Year = 0 then D.Year := Temp.Year else
    if D.Year < 100 then D.Year := Temp.Year div 100 * 100 + D.Year;
  if Within(D.Month, 1, 12) and                    {  । 1..12}
    (Within(D.Day, 1, SizeOfMonth[D.Month]) or     {   । 1..31}
    ((D.Year mod 4 = 0) and Within(D.Day, 1, SizeOfMonth[D.Month]+1)))
    then Exit;
  for i := 1 to 3 do Arr(D)[i] := 0;
end;

function DtoC(D: Date): string;
var
  St: string[8];
begin
  St := DtoS(D);
  if St <> Space(8) then St := Copy(St, 7, 2) +
    Delimiter + Copy(St, 5, 2) + Delimiter + Copy(St, 3, 2);
  DtoC := St;
end;

function Empty(D: Date): Boolean;
begin
  Empty := (D.Year = 0) or (D.Month = 0) or (D.Day = 0)
end;

function DtoS(D: Date): string;
var
  Yy: string[4];
  Mm, Dd: string[2];
begin
  DtoS := Space(8);
  if Empty(D) then Exit;
  Str(D.Year, Yy);  while Length(Yy) <> 4 do Insert('0', Yy, 1);
  Str(D.Month, Mm); while Length(Mm) <> 2 do Insert('0', Mm, 1);
  Str(D.Day, Dd);   while Length(Dd) <> 2 do Insert('0', Dd, 1);
  DtoS := Yy + Mm + Dd;
end;

{==========================================================================}

constructor TBase.Init(FName: string);
label
  Deny;
var
  Result, Count,
  Offset: Word;
  Rec: PField;
  Dir: DirStr;
  Name: NameStr;
  Ext: ExtStr;
begin
  inherited Init;
  FSplit(FName, Dir, Name, Ext);
  if Ext = '' then Ext := '.DBF';
  Assign(Dbf, FExpand(Dir + Name + Ext));
  {$I-}  Reset(Dbf, 1);  {$I+}
  if IOResult <> 0 then Goto Deny;
  BlockRead(Dbf, Dbf_Id, BlockLen, Result);
  if (Result <> BlockLen) or
    not (Dbf_Id in [DataBase, MemoFile]) then Goto Deny;
  Count := (Header - BlockLen) div BlockLen;
  Fields := New(PFieldsCollection, Init(Count, 0));
  Offset := 1;
  for Count := 1 to Count do
  begin
    Rec := MAlloc(SizeOf(TField));
    if Rec = nil then Goto Deny;
    BlockRead(Dbf, Rec^, BlockLen, Result);
    if Result <> BlockLen then Goto Deny;
    Rec^.Addr := Offset;
    Inc(Offset, Rec^.Size);
    Fields^.Insert(Rec);
  end;
  Buffer := MAlloc(RecSize + 1);
  if Buffer = nil then Goto Deny;
  if Dbf_Id = MemoFile then
  begin
    Ext := '.DBT';
    Assign(Dbt, FExpand(Dir + Name + Ext));
    {$I-}  Reset(Dbt, 1);  {$I+}
    if IOResult <> 0 then Goto Deny;
  end;
  GoRec(1);
  Exit;
Deny: {⪠  樠樨}
  Done;
  Fail;
end;

destructor TBase.Done;
var
  Result: Integer;
begin
  Update;
  if Updated then
  begin
    Seek(Dbf, Header + LastRec * RecSize);
    BlockWrite(Dbf, MarkEnd, SizeOf(Byte));
    Truncate(Dbf);
  end;
  if Buffer <> nil then FreeMem(Buffer, RecSize + 1);
  if Fields <> nil then Dispose(Fields, Done);
{$I-}
  Close(Dbt);  Result := IOResult;
  Close(Dbf);  Result := IOResult;
{$I+}
  inherited Done;
end;

procedure TBase.GoRec(No: Longint);
begin
  Commit;
  if (No > LastRec) or (No < 1) then
  begin
    RecNo := LastRec + 1;
    FillChar(Buffer^, RecSize, ' ');
  end else
  begin
    RecNo := No;
    Seek(Dbf, Header + (Pred(RecNo)) * RecSize);
    BlockRead(Dbf, Buffer^, RecSize);
  end;
  Eof := RecNo > LastRec;
end;

procedure TBase.Commit;
begin
  if not Modified then Exit;
  Modified := False;  Updated := True;
  Seek(Dbf, Header + (Pred(RecNo)) * RecSize);
  BlockWrite(Dbf, Buffer^, RecSize);
end;

procedure TBase.Update;
  procedure WriteField(Rec: PField); far;
  begin
    BlockWrite(Dbf, Rec^, BlockLen);
  end;
var
  Year, Month, Day, DayOfWeek: Word;
begin
  Commit;  if not Updated then Exit;
  GetDate(Year, Month, Day, DayOfWeek);
  LUpDate.Year := Year mod 100;
  LUpDate.Month := Month;
  LUpDate.Day := Day;
  Seek(Dbf, 0);
  BlockWrite(Dbf, Dbf_Id, BlockLen);
  Fields^.ForEach(@WriteField);
  BlockWrite(Dbf, MarkBeg, SizeOf(Byte));
end;

procedure TBase.Append;
begin
  GoRec(0);
  Modified := True;
  Commit;
  Inc(LastRec);
end;

function TBase.FCount: Integer;
begin
  FCount := Fields^.Count;
end;

function TBase.At(Key: KeyStr): PField;
  function Match(P: PField): Boolean; far;
  begin
    Match := StrPas(P^.Name) = Key;
  end;
begin
  At := Fields^.FirstThat(@Match);
end;

function TBase.Col(No: Integer): PField;
begin
  Col := Fields^.At(Pred(No));
end;

function TBase.Alias: string;
begin
  Alias := StrPas(FileRec(Dbf).Name);
end;

function TBase.Get(Key: KeyStr): string;
var
  F: PField;
  St: string;
  Dest: array[0..255] of Char;
  Num: Real;
  Code: Integer;
begin
  Get := ''; F := At(Key); if F = nil then Exit;
  St := StrPas(StrLCopy(Dest, Buffer + F^.Addr, F^.Size));
  case F^.Flag of
    'N': begin
           Val(St, Num, Code);
           if Num = 0 then St := Space(F^.Size);
         end;
    'D': if St <> Space(8) then St := Copy(St, 7, 2) +
           Delimiter + Copy(St, 5, 2) + Delimiter + Copy(St, 3, 2);
    'L': if St <> 'T' then St := ' ';
    'M': St := ' < Memo > ';
  end;
  Get := St;
end;

procedure TBase.Put(Key: KeyStr; var Info: string);
var
  F: PField;
  Num: Real;
  Code: Integer;
  Dest: array[0..255] of Char;
  D: DATE;
begin
  F := At(Key);  if F = nil then Exit;
  case F^.Flag of
    'C': while Length(Info) < F^.Size do Info := Info + ' ';
    'N': begin
           Val(Info, Num, Code);
           Str(Num: F^.Size: F^.Dec, Info);
           if (Code <> 0) or (Length(Info) > F^.Size) then
             Str(0.0: F^.Size: F^.Dec, Info);
         end;
    'D': begin CtoD(Info, D); Info := DtoS(D) end;
    'L': if (Info <> '') and (UpCase(Info[1]) in ['T','Y'])
           then Info := 'T' else Info := 'F';
    'M': Exit;
  end;
  StrMove(Buffer + F^.Addr, StrPCopy(Dest, Info), F^.Size);
  Modified := True;
end;

end.